import logging
import os
import signal
import webbrowser
from time import sleep

import streamlit as st

import utils
from langchain_coder import LangChainCoder

# logging.basicConfig(level = logging.DEBUG,format='%(levelname)s-%(message)s')

try:
    from dotenv import load_dotenv

    load_dotenv()
except Exception as e:
    logging.error("dotenv import error but no needed")

num_of_iterations = 10


def generate_response(txt, title):
    """
    Generate response using the LangChainCoder.

    Args:
        txt (str): The input text.

    Yields:
        dict: A dictionary containing response information.
    """
    for data in agent(txt, title, num_of_iterations):
        yield data


# Page title
title = "🧩 DemoGPT"

st.set_page_config(page_title=title)
st.title(title)
# Text input

openai_api_key = st.sidebar.text_input(
    "OpenAI API Key",
    placeholder="sk-...",
    value=os.getenv("OPENAI_API_KEY", ""),
    type="password",
)
demo_title = st.text_input("Enter your demo title", placeholder="Type your demo title")
empty_idea = st.empty()
demo_idea = empty_idea.text_area(
    "Enter your LLM-based demo idea", placeholder="Type your demo idea here", height=100
)

st.write("Examples")

cols = st.columns([1, 1, 1.2])

PROGRESS_BAR_INFO = {
    "start": {"text": "Generating Code...", "percentage": 25},
    "langchain": {"text": "Transforming to streamlit code", "percentage": 35},
    "refining": {"text": "Refining Code...", "percentage": 50},
    "streamlit": {"text": "Creating App...", "percentage": 75},
    "failed": {"text": "Failed", "percentage": 100},
}


def progressBar(key, bar=None):
    info = PROGRESS_BAR_INFO[key]
    if bar:
        bar.progress(info["percentage"], text=info["text"])
    else:
        return st.progress(info["percentage"], text=info["text"])


examples = [
    "Language Translator 📝",
    "Grammer Corrector 🛠",
    "Blog post generator from title 📔",
]

if "pid" not in st.session_state:
    st.session_state["pid"] = -1

example_submitted = False

with st.form("a", clear_on_submit=True):
    submitted = st.form_submit_button("Submit")
    for col, example in zip(cols, examples):
        if col.button(example):
            example_submitted = True
            demo_idea = empty_idea.text_area(
                "Enter your LLM-based demo idea", example, height=100
            )
            logging.info(f"Demo Idea:{demo_idea}")

    if submitted or example_submitted:

        if not openai_api_key.startswith("sk-"):
            st.warning("Please enter your OpenAI API Key!", icon="⚠️")
        else:
            bar = progressBar("start")

            agent = LangChainCoder(openai_api_key=openai_api_key)

            if st.session_state["pid"] != -1:
                logging.info(f"Terminating the previous applicaton ...")
                os.kill(st.session_state["pid"], signal.SIGTERM)
                st.session_state["pid"] = -1

            code_empty = st.empty()
            for data in generate_response(demo_idea, demo_title):
                code = data["code"]
                success = data["success"]
                task_id = data["task_id"]
                stage = data["stage"]

                if success:
                    progressBar(stage, bar)
                    if stage == "streamlit" and task_id == "final":
                        with st.expander("Code"):
                            st.code(code)
                        example_submitted = False
                        st.session_state["pid"] = utils.runStreamlit(
                            code, openai_api_key
                        )
                        sleep(5)
                        webbrowser.open("http://localhost:8502")
                        break
                else:
                    progressBar("refining")
            else:
                progressBar("failed")



import os
import re
from pprint import pprint
import json
os.environ["OPENAI_API_KEY"]="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"

from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)

llm = ChatOpenAI(model="gpt-3.5-turbo", temperature=0)
def getRes(system_template="", human_template="", **kwargs):
    prompts = []
    if system_template:
        prompts.append(SystemMessagePromptTemplate.from_template(system_template))
    if human_template:
        prompts.append(HumanMessagePromptTemplate.from_template(human_template))
    chat_prompt = ChatPromptTemplate.from_messages(prompts)
    return LLMChain(llm=llm, prompt=chat_prompt).run(**kwargs)

examples = """
=========================================
Instruction:Generate blog post from title.
Args: {{
"system_template":"You are a helpful assistant that generates a blog post from the title: {{title}}. Please provide some content.",
"template":"{{title}}",
"variety":"True"
}}
##########################################
Instruction:Implement language translation app
Args: {{
"system_template":"You are a helpful assistant that translates {{input_language}} to {{output_language}}.",
"template":"{{text}}",
"variety":"False"
}}
##########################################
Instruction:Generate animal name from animal
Args: {{
"system_template":"You are a helpful assistant that generates a name for an animal. You generate short answer.",
"template":"What is a good name for a {{animal}}?",
"variety":"True"
}}
##########################################
Instruction:Create programming related humor machine
Args: {{
"system_template":"You are a helpful assistant that generates a humor related to programming.",
"template":"",
"variety":"True"
}}
##########################################
Instruction:Create a math teacher.
Args: {{
"system_template":"You are a helpful assistant that solve any math problem",
"template":"Here is the math problem: {{math_problem}}",
"variety":"False"
}}
##########################################
Instruction:{instruction}
Args:
"""

stremalit_system_template = """
Transfer the python code to the interactive streamlit code to do the instruction.
Make the output fancy.
Rewrite the python code to generate full streamlit code
"""
stremalit_human_template="""
Instruction:{instruction}
Python Code:{code}
Streamlit Code:
"""

def getCode(
instruction = "Create an app that predicts the gender from country and name inforation"
):
    res = getRes(human_template=examples,instruction=instruction)
    templates = json.loads(res)

    variables = [var[1:-1] for var in re.findall("{\S+}", res)]
    args = ", ".join(variables)
    run_call = '{}'
    if len(variables) > 0:
        run_call = ", ".join([f"{var}={var}" for var in variables])
    
    temperature = 0 if templates.get("variety","False") == "False" else 0.7

    code =f"""
    from langchain import LLMChain
    from langchain.chat_models import ChatOpenAI
    from langchain.prompts.chat import (ChatPromptTemplate,
                                        HumanMessagePromptTemplate,
                                        SystemMessagePromptTemplate)

    def generate({args}):
        chat = ChatOpenAI(
            temperature={temperature}
        )
        system_template = {templates['system_template']}
        system_message_prompt = SystemMessagePromptTemplate.from_template(template)
        human_template = {templates['template']}
        human_message_prompt = HumanMessagePromptTemplate.from_template(human_template)
        chat_prompt = ChatPromptTemplate.from_messages(
            [system_message_prompt, human_message_prompt]
        )

        chain = LLMChain(llm=chat, prompt=chat_prompt)
        result = chain.run({run_call})
        return result                                 
    """

    streamlit_code = getRes(system_template = stremalit_system_template,
                            human_template=stremalit_human_template,
                            instruction=instruction,
                            code=code)
    return streamlit_code

__all__ = ['getCode']


import logging
import os

import fire
import langchain
from langchain.chat_models import ChatOpenAI
from langchain.docstore.document import Document
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.vectorstores import Chroma
from tqdm import tqdm

import utils
from chains.chains import Chains
from langchain_expert import LangChainExpert


class LangChainCoder:
    def __init__(
        self,
        model_name="gpt-3.5-turbo-16k",
        persist_directory="langchain_code",
        device="cuda",
        distance_metric="cos",
        maximal_marginal_relevance=True,
        k=4,
    ):
        self.root_dir = "/".join(langchain.__file__.split("/")[:-1])
        self.model_name = model_name
        self.persist_directory = persist_directory
        self.device = device
        self.distance_metric = distance_metric
        self.maximal_marginal_relevance = maximal_marginal_relevance
        self.k = k
        self.embeddings = HuggingFaceEmbeddings(model_kwargs={"device": device})
        self.__initModels()

    def __constructDB(self):
        if os.path.exists(self.persist_directory):
            logging.info("DB has been found and fetched")
            self.db = Chroma(
                persist_directory=self.persist_directory,
                embedding_function=self.embeddings,
            )
        else:
            logging.info("DB not found, creation started...")
            docs = []
            for dirpath, _, filenames in tqdm(os.walk("examples/goals/")):
                for file in filenames:
                    if file.endswith(".md"):
                        filepath = os.path.join(dirpath, file)
                        source_path = filepath.replace("goals", "codes/").replace(
                            ".md", ".py"
                        )
                        with open(source_path) as sf:
                            metadata = {"source": sf.read()}
                        with open(filepath) as f:
                            docs.append(
                                Document(page_content=f.read(), metadata=metadata)
                            )
            self.db = Chroma.from_documents(
                docs, self.embeddings, persist_directory=self.persist_directory
            )

    def __constructRetriever(self):
        self.__constructDB()
        self.retriever = self.db.as_retriever()
        self.retriever.search_kwargs["distance_metric"] = self.distance_metric
        self.retriever.search_kwargs[
            "maximal_marginal_relevance"
        ] = self.maximal_marginal_relevance
        self.retriever.search_kwargs["k"] = self.k
        logging.info("Retriever has been succesfully constructed")

    def __initModels(self):
        self.__constructRetriever()
        self.model = ChatOpenAI(model_name=self.model_name, temperature=0)
        self.expert = LangChainExpert()

    def __getSource(self, query):
        resulting_text = ""
        docs = self.retriever.get_relevant_documents(query)
        for doc in docs:
            resulting_text += doc.metadata["source"]
            resulting_text += "\n" + "#" * 40
        return resulting_text

    def __getTasks(self, task):
        instruction_list = Chains.divide(task=task)
        if instruction_list.startswith("["):
            instruction_list = instruction_list[1:]
        if instruction_list.endswith("]"):
            instruction_list = instruction_list[:-1]
        return [insruction.strip() for insruction in instruction_list.split(",")]

    def __mergeTasks(self, task, examples):
        merged_code = Chains.merge(task=task, examples=examples)
        merged_code = utils.refineCode(merged_code)
        return merged_code

    def __getSubResult(self, query, iterations):
        doc = self.__getSource(query)
        draft_code = Chains.draft(document=doc, idea=query)
        for _ in range(iterations):
            _, error, success = utils.runPython(draft_code)
            if not success:
                feedback = self.expert.debug(error)
                draft_code = Chains.debug(
                    draft_code=draft_code, idea=query, feedback=feedback, document=doc
                )
            else:
                return utils.refineCode(draft_code)
        return None

    def __getLangChainCode(self, instruction, iterations):
        tasks = self.__getTasks(instruction)
        print(tasks)
        examples = ""
        for i, task in enumerate(tasks):
            code = self.__getSubResult(task, iterations=iterations)
            if code:
                examples += f"""
                Subtask {i+1} : {task}
                Code {i+1} : {code}
                {'#'*40}

                """
        if len(tasks) == 1:
            return code
        final_code = self.__mergeTasks(instruction, examples)
        return final_code

    def __getStreamlitCode(self, instruction, langchain_code, title):
        merged_code = Chains.streamlit(
            instruction=instruction, langchain_code=langchain_code, title=title
        )
        merged_code = utils.refineCode(merged_code)
        return merged_code

    def code(
        self,
        instruction="Create a translation system that converts English to French",
        title="my translator",
        iterations=10,
    ):
        langchain_code = self.__getLangChainCode(instruction, iterations=iterations)
        return self.__getStreamlitCode(instruction, langchain_code, title)


if __name__ == "__main__":
    coder = LangChainCoder()
    fire.Fire(coder.code)



# cli.py

import subprocess
import os

def main():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    subprocess.run(["streamlit", "run", os.path.join(current_dir, "app.py")])

if __name__ == "__main__":
    main()



import logging
import os

import fire
import langchain
from langchain.docstore.document import Document
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.vectorstores import Chroma
from termcolor import colored
from tqdm import tqdm

import utils
from chains.chains import Chains
from langchain_expert import LangChainExpert


class LangChainCoder:
    def __init__(
        self,
        openai_api_key="sk-",
        model_name="gpt-3.5-turbo-16k",
        data_root = "src/data_alpha/",
        persist_directory="langchain_code",
        device="cuda",
        distance_metric="cos",
        maximal_marginal_relevance=True,
        k=2,
    ):
        self.root_dir = "/".join(langchain.__file__.split("/")[:-1])
        Chains.setLlm(model_name, openai_api_key)
        self.persist_directory = os.path.join(data_root,persist_directory)
        self.goals_directory = os.path.join(data_root,"examples/goals/")
        self.device = device
        self.distance_metric = distance_metric
        self.maximal_marginal_relevance = maximal_marginal_relevance
        self.k = k
        self.embeddings = HuggingFaceEmbeddings(model_kwargs={"device": device})
        self.__initModels()

    def __constructDB(self):
        if os.path.exists(self.persist_directory):
            logging.info("DB has been found and fetched")
            self.db = Chroma(
                persist_directory=self.persist_directory,
                embedding_function=self.embeddings,
            )
        else:
            logging.info("DB not found, creation started...")
            docs = []
            for dirpath, _, filenames in tqdm(os.walk(self.goals_directory)):
                for file in filenames:
                    if file.endswith(".md"):
                        filepath = os.path.join(dirpath, file)
                        source_path = filepath.replace("goals", "codes/").replace(
                            ".md", ".py"
                        )
                        with open(source_path) as sf:
                            metadata = {"source": sf.read()}
                        with open(filepath) as f:
                            docs.append(
                                Document(page_content=f.read(), metadata=metadata)
                            )
            self.db = Chroma.from_documents(
                docs, self.embeddings, persist_directory=self.persist_directory
            )

    def __constructRetriever(self):
        self.__constructDB()
        self.retriever = self.db.as_retriever()
        self.retriever.search_kwargs["distance_metric"] = self.distance_metric
        self.retriever.search_kwargs[
            "maximal_marginal_relevance"
        ] = self.maximal_marginal_relevance
        self.retriever.search_kwargs["k"] = self.k
        logging.info("Retriever has been succesfully constructed")

    def __initModels(self):
        self.__constructRetriever()
        self.expert = LangChainExpert()

    def __getSource(self, query):
        resulting_text = ""
        docs = self.retriever.get_relevant_documents(query)
        for doc in docs:
            resulting_text += doc.metadata["source"]
            resulting_text += "\n" + "#" * 40
        return resulting_text

    def __getTasks(self, task):
        instruction_list = Chains.divide(task=task)
        if instruction_list.startswith("["):
            instruction_list = instruction_list[1:]
        if instruction_list.endswith("]"):
            instruction_list = instruction_list[:-1]
        return [insruction.strip() for insruction in instruction_list.split(",")]

    def __mergeTasks(self, task, examples):
        merged_code = Chains.merge(task=task, examples=examples)
        merged_code = utils.refineCode(merged_code)
        return merged_code

    def __getSubResult(self, query, doc, index, iterations):
        draft_code = Chains.draft(document=doc, idea=query)
        for res in self.__solveBugs(draft_code, query, doc, index + 1, iterations):
            yield res
            success = res["success"]
            if success:
                break
        yield res

    def __solveBugs(self, draft_code, query, doc, index, iterations):
        for i in range(iterations):
            _, error, success = utils.runPython(draft_code)
            if not success:
                feedback = self.expert.debug(error)
                print(colored(feedback, "red"))
                draft_code = Chains.debug(
                    draft_code=draft_code, idea=query, feedback=feedback, document=doc
                )
                draft_code = utils.refineCode(draft_code)
                yield {
                    "task_id": index,
                    "progress": f"{i+1}/{iterations}",
                    "success": False,
                    "code": draft_code,
                    "stage": "langchain",
                }
            else:
                draft_code = utils.refineCode(draft_code)
                yield {
                    "task_id": index,
                    "progress": f"{i+1}/{iterations}",
                    "success": True,
                    "code": draft_code,
                    "stage": "langchain",
                }

                break
        else:
            yield {
                "task_id": index + 1,
                "progress": f"{iterations}/{iterations}",
                "success": False,
                "code": draft_code,
                "stage": "langchain",
            }

    def __getLangChainCode(self, instruction, iterations):
        tasks = self.__getTasks(instruction)
        print(colored(tasks,"yellow"))
        print(colored("Tasks have been generated", "green"))
        examples = ""
        for i, task in enumerate(tasks):
            doc = self.__getSource(task)
            for res in self.__getSubResult(task, doc, i, iterations=iterations):
                text = f"Task[{res['task_id']}/{len(tasks)}]\nProgress: {res['progress']}\nSuccess: {res['success']}\n"
                print(task)
                print(colored(res["code"],"blue"))
                if res["success"]:
                    print(colored(text, "green"))
                else:
                    print(colored(text, "yellow"))
                if res["success"]:
                    code = res["code"]
                    examples += f"""
                    Subtask {i+1} : {task}
                    Code {i+1} : {code}
                    {'#'*40}

                    """
                    break
                yield res

        if len(tasks) == 1:
            yield res
        else:
            final_code = self.__mergeTasks(instruction, examples)
            doc = self.__getSource(instruction)
            for res in self.__solveBugs(
                final_code, instruction, doc, "merge", iterations
            ):
                success = res["success"]
                text = f"Task[{res['task_id']}/{len(tasks)}]\nProgress: {res['progress']}\nSuccess: {res['success']}\n"
                if success:
                    print(colored(text, "green"))
                else:
                    print(colored(text, "yellow"))
                if success:
                    final_code = res["code"]
                    break
                yield res
            print(colored("Tasks have been merged", "green"))

    def __getStreamlitCode(self, instruction, langchain_code, title):
        merged_code = Chains.streamlit(
            instruction=instruction, langchain_code=langchain_code, title=title
        )
        merged_code = utils.refineCode(merged_code)
        return {
            "code": merged_code,
            "stage": "streamlit",
            "success": True,
            "task_id": "draft",
        }

    def __getFinalCode(self, instruction, code):
        feedback = Chains.feedback(code=code)
        code = Chains.refine(instruction=instruction, code=code, feedback=feedback)
        return utils.refineCode(code)

    def __getFinalCode1(self, instruction, code):
        code = Chains.refine1(code=code)
        return utils.refineCode(code)

    def __call__(
        self,
        instruction="Create a translation system that converts English to French",
        title="my translator",
        iterations=10,
    ):
        for res in self.__getLangChainCode(instruction, iterations=iterations):
            yield res
        langchain_code = res["code"]
        res = self.__getStreamlitCode(instruction, langchain_code, title)
        yield res
        streamlit_code = res["code"]
        print(colored(streamlit_code, "light_green"))
        final_code = self.__getFinalCode1(instruction, streamlit_code)

        yield {
            "code": final_code,
            "success": True,
            "task_id": "final",
            "stage": "streamlit",
        }


if __name__ == "__main__":
    coder = LangChainCoder()
    fire.Fire(coder.code)



import logging
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import threading
from subprocess import PIPE
from threading import Timer

from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI

from prompt_based.prompts import *
from pkg_resources import resource_stream


class BaseModel:
    """
    Base class for the LogicModel and StreamlitModel classes.

    Methods:
        - __init__(self, openai_api_key: str):
            Initializes the BaseModel with the provided OpenAI API key.

        - refine_code(self, code: str) -> str:
            Refines the provided code by removing unnecessary parts.
    """

    def __init__(self, openai_api_key):
        """
        Initializes the BaseModel with the provided OpenAI API key.

        Args:
            openai_api_key (str): The OpenAI API key.
        """
        self.openai_api_key = openai_api_key
        self.llm = ChatOpenAI(openai_api_key=openai_api_key, temperature=0.0)

    def refine_code(self, code):
        """
        Refines the provided code by removing unnecessary parts.

        Args:
            code (str): The code to be refined.

        Returns:
            str: The refined code.
        """
        if "```" in code:
            code = code.split("```")[1]
            if code.startswith("python"):
                code = code[len("python") :].strip()
        return code


class LogicModel(BaseModel):
    """
    Represents the logic model for generating Python code.

    Methods:
        - __init__(self, openai_api_key: str):
            Initializes the LogicModel with the provided OpenAI API key.

        - addDocuments(self):
            Adds documents to the logic model for generating Python code.

        - decode_results(self, results) -> Generator[str, None, None]:
            Decodes the results returned by the GPT-3.5-turbo model.

        - run_python(self, code: str) -> Tuple[str, str, bool]:
            Executes the Python code using the subprocess module.

        - __call__(self, topic: str, num_iterations: int = 10) -> Generator[Dict[str, Any], None, None]:
            Executes the logic model to generate Python code based on the given topic.
    """

    def __init__(self, openai_api_key):
        """
        Initializes the LogicModel with the provided OpenAI API key.

        Args:
            openai_api_key (str): The OpenAI API key.
        """
        super().__init__(openai_api_key)
        self.code_chain = LLMChain(llm=self.llm, prompt=code_prompt)
        self.test_chain = LLMChain(llm=self.llm, prompt=test_prompt)
        self.refine_chain = LLMChain(llm=self.llm, prompt=refine_chat_prompt)
        self.fix_chain = LLMChain(llm=self.llm, prompt=fix_chat_prompt)
        self.check_chain = LLMChain(llm=self.llm, prompt=check_chat_prompt)
        self.addDocuments()

    def addDocuments(self):
        """
        Adds documents to the logic model for generating Python code.
        """
        self.document = ""
        for path in ["prompts.txt"]:
            with resource_stream('prompt_based', path) as f:
                self.document += f.read().decode('utf-8')

    def decode_results(self, results):
        """
        Decodes the results returned by the executed python code

        Args:
            results: The results to be decoded.

        Returns:
            Tuple[str, str]: The decoded results.
        """
        return (res.strip().decode("utf-8", errors="ignore") for res in results)

    def run_python(self, code):
        """
        Executes the Python code using the subprocess module.

        Args:
            code (str): The Python code to be executed.

        Returns:
            Tuple[str, str, bool]: The output, error, and success status of the execution.
        """
        tmp = tempfile.NamedTemporaryFile("w", suffix=".py", delete=False, encoding='utf-8')
        tmp.write(code)
        tmp.flush()
        environmental_variables = {"OPENAI_API_KEY": self.openai_api_key}
        python_path = shutil.which("python")
        if platform.system() == "Windows":
            env = os.environ.copy()
            env["PYTHONPATH"] = ""
            env["OPENAI_API_KEY"] = self.openai_api_key
            python_path = sys.executable
            process = subprocess.Popen(
                [python_path, tmp.name], env=env, stdout=PIPE, stderr=PIPE
            )
        else:
            process = subprocess.Popen(
                [python_path, tmp.name],
                env=environmental_variables,
                stdout=PIPE,
                stderr=PIPE,
            )
        try:
            tmp.close()
        except PermissionError:
            pass
        output, err = self.decode_results(process.communicate())
        success = len(err) == 0
        return output, err, success

    def __call__(self, topic, num_iterations=10):
        """
        Executes the logic model to generate Python code based on the given topic.

        Args:
            topic (str): The topic or goal of the code generation.
            num_iterations (int, optional): The maximum number of iterations. Defaults to 10.

        Yields:
            Generator[Dict[str, Any], None, None]: A dictionary containing information about each iteration.
        """
        error = ""
        percentage = 0
        feedback = ""
        total_code = ""
        refined_code = ""

        for i in range(num_iterations):
            if error:
                code = self.refine_chain.run(
                    content=code, critics=feedback, document=self.document
                )
            else:
                code = self.code_chain.run(document=self.document, topic=topic)
            code = self.refine_code(code)

            test_code = self.test_chain.run(code=code, topic=topic, feedback=feedback)
            test_code = self.refine_code(test_code)

            total_code = code + "\n" + test_code

            response, error, success = self.run_python(total_code)

            if not success:
                logging.warning(f"Iteration:{i}:{error}")

            if success:
                break

            response = error

            feedback = self.fix_chain.run(code=total_code, error=error)

            percentage += 100 // num_iterations

            yield {
                "code": code,
                "total_code": total_code,
                "success": success,
                "out": response,
                "error": error,
                "feedback": feedback,
                "test_code": test_code,
                "refined_code": refined_code,
                "percentage": min(100, percentage),
            }

        yield {
            "code": code,
            "total_code": total_code,
            "success": success,
            "out": response,
            "error": error,
            "feedback": feedback,
            "test_code": test_code,
            "refined_code": refined_code,
            "percentage": 100,
        }


class StreamlitModel(BaseModel):
    """
    Represents the Streamlit model for generating Streamlit applications.

    Methods:
        - __init__(self, openai_api_key: str):
            Initializes the StreamlitModel with the provided OpenAI API key.

        - run_code(self, code: str) -> int:
            Runs the provided code as a Streamlit application and returns the process ID.

        - __call__(self, topic: str, title: str, code: str, test_code: str, progress_func: Callable[[int, str]], success_func: Callable[[], None]) -> int:
            Executes the Streamlit model to generate a Streamlit application.
    """

    def __init__(self, openai_api_key):
        """
        Initializes the StreamlitModel with the provided OpenAI API key.

        Args:
            openai_api_key (str): The OpenAI API key.
        """
        super().__init__(openai_api_key)
        self.streamlit_code_chain = LLMChain(llm=self.llm, prompt=streamlit_code_prompt)

    def runThread(self, proc):
        proc.communicate()

    def run_code(self, code):
        """
        Runs the provided code as a Streamlit application and returns the process ID.

        Args:
            code (str): The code of the Streamlit application.

        Returns:
            int: The process ID of the Streamlit application.
        """
        tmp = tempfile.NamedTemporaryFile("w", suffix=".py", delete=False,encoding='utf-8')
        tmp.write(code)
        tmp.flush()
        environmental_variables = {
            "OPENAI_API_KEY": self.openai_api_key,
            "STREAMLIT_SERVER_PORT": "8502",
        }
        streamlit_path = shutil.which("streamlit")
        if platform.system() == "Windows":
            env = os.environ.copy()
            env["PYTHONPATH"] = ""
            env["OPENAI_API_KEY"] = self.openai_api_key
            env["STREAMLIT_SERVER_PORT"] = "8502"
            python_path = sys.executable
            process = subprocess.Popen(
                [python_path, "-m", "streamlit", "run", tmp.name],
                env=env,
                stdout=PIPE,
                stderr=PIPE,
            )
            threading.Thread(target=self.runThread, args=(process,)).start()
        else:
            process = subprocess.Popen(
                [streamlit_path, "run", tmp.name],
                env=environmental_variables,
                stdout=PIPE,
                stderr=PIPE,
            )
        try:
            tmp.close()
        except PermissionError:
            pass

        return process.pid

    def __call__(self, topic, title, code, test_code, progress_func, success_func):
        """
        Executes the Streamlit model to generate a Streamlit application.

        Args:
            topic (str): The topic or goal of the application.
            title (str): The title of the Streamlit application.
            code (str): The logic code of the application.
            test_code (str): The test code of the application.
            progress_func (Callable[[int, str]]): A function to update the progress of the application generation.
            success_func (Callable[[], None]): A function to indicate the success of the application generation.

        Returns:
            int: The process ID of the Streamlit application.
        """
        streamlit_code = self.streamlit_code_chain.run(
            topic=topic, title=title, logic_code=code, test_code=test_code
        )
        refined_code = self.refine_code(streamlit_code)
        progress_func(100, "Redirecting to the demo page...")
        success_func()
        return self.run_code(refined_code), refined_code



import os
import shutil
import subprocess
import sys
import tempfile
from subprocess import PIPE, TimeoutExpired

import fire
from dotenv import load_dotenv
from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.docstore.document import Document
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.vectorstores import Chroma
from termcolor import colored
from tqdm import tqdm

from agent_prompts import *
from langchain_expert import LangChainExpert

load_dotenv()

from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)


def decodeResults(results):
    return (res.strip().decode("utf-8") for res in results)


def refineCode(code):
    if "```" in code:
        code = code.split("```")[1]
        if code.startswith("python"):
            code = code[len("python") :].strip()
    return code


def runPython(code, timeout_sec=10):
    code = refineCode(code)
    with tempfile.NamedTemporaryFile("w") as tmp:
        tmp.write(code)
        tmp.flush()
        environmental_variables = {"OPENAI_API_KEY": os.getenv("OPENAI_API_KEY")}
        python_path = shutil.which("python")
        if not python_path:  # shows 'which' returns None
            python_path = sys.executable
        process = subprocess.Popen(
            [python_path, tmp.name],
            env=environmental_variables,
            stdout=PIPE,
            stderr=PIPE,
        )
        try:
            output, err = decodeResults(process.communicate(timeout=timeout_sec))
            success = len(err) == 0
        except TimeoutExpired:
            process.kill()
            success = True
            output = err = ""
        return output, err, success


persist_directory = "goals_db"
embeddings = HuggingFaceEmbeddings(model_kwargs={"device": "cuda"})
if os.path.exists(persist_directory):
    db = Chroma(persist_directory=persist_directory, embedding_function=embeddings)
else:
    docs = []
    for dirpath, dirnames, filenames in tqdm(os.walk("examples/goals/")):
        for file in filenames:
            if file.endswith(".md"):
                filepath = os.path.join(dirpath, file)
                source_path = filepath.replace("goals", "codes/").replace(".md", ".py")
                with open(source_path) as sf:
                    metadata = {"source": sf.read()}
                with open(filepath) as f:
                    docs.append(Document(page_content=f.read(), metadata=metadata))
    db = Chroma.from_documents(docs, embeddings, persist_directory=persist_directory)

retriever = db.as_retriever()
retriever.search_kwargs["distance_metric"] = "cos"
retriever.search_kwargs["maximal_marginal_relevance"] = True
retriever.search_kwargs["k"] = 4

llm = ChatOpenAI(model="gpt-3.5-turbo-16k", temperature=0)

expert = LangChainExpert()


def getSource(query):
    resulting_text = ""
    docs = retriever.get_relevant_documents(query)
    for doc in docs:
        resulting_text += doc.metadata["source"]
        resulting_text += "\n" + "#" * 40
    return resulting_text


def getTasks(task):
    system_message_prompt = SystemMessagePromptTemplate.from_template(
        DIVIDE_TASKS_SYSTEM_TEMPLATE
    )
    human_message_prompt = HumanMessagePromptTemplate.from_template(
        DIVIDE_TASKS_HUMAN_TEMPLATE
    )
    chat_prompt = ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )
    chain = LLMChain(llm=llm, prompt=chat_prompt)
    instruction_list = chain.run(task=task)
    if instruction_list.startswith("["):
        instruction_list = instruction_list[1:]
    if instruction_list.endswith("]"):
        instruction_list = instruction_list[:-1]
    return [insruction.strip() for insruction in instruction_list.split(",")]


def mergeTasks(task, examples):
    system_message_prompt = SystemMessagePromptTemplate.from_template(
        MERGE_CODES_SYSTEM_TEMPLATE
    )
    human_message_prompt = HumanMessagePromptTemplate.from_template(
        MERGE_CODES_HUMAN_TEMPLATE
    )
    chat_prompt = ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )
    chain = LLMChain(llm=llm, prompt=chat_prompt)
    merged_code = chain.run(task=task, examples=examples)
    merged_code = refineCode(merged_code)
    return merged_code


def getSubResult(query, iterations):
    doc = getSource(query)

    human_message_prompt = HumanMessagePromptTemplate.from_template(DOC_USE_TEMPLATE)
    chat_prompt = ChatPromptTemplate.from_messages([human_message_prompt])
    chain = LLMChain(llm=llm, prompt=chat_prompt)

    draft_code = chain.run(document=doc, idea=query)

    for _ in range(iterations):
        print(colored(draft_code, "yellow"))
        output, error, success = runPython(draft_code)
        print(colored(output, "blue"))
        print(colored(error, "red"))
        if not success:
            question = f"""
            How to solve that error below?

            {error}
            """

            feedback = expert.ask(question)

            print(colored(feedback, "blue"))

            human_message_prompt = HumanMessagePromptTemplate.from_template(
                APP_DEBUGGING_TEMPLATE
            )
            chat_prompt = ChatPromptTemplate.from_messages([human_message_prompt])
            chain = LLMChain(llm=llm, prompt=chat_prompt)
            draft_code = chain.run(
                draft_code=draft_code, idea=query, feedback=feedback, document=doc
            )
        else:
            return refineCode(draft_code)
    return None


def getLangChainCode(instruction, iterations):
    tasks = getTasks(instruction)
    print(tasks)
    examples = ""
    for i, task in enumerate(tasks):
        code = getSubResult(task, iterations=iterations)
        if code:
            examples += f"""
            Subtask {i+1} : {task}
            Code {i+1} : {code}
            {'#'*40}

            """
    if len(tasks) == 1:
        return code
    final_code = mergeTasks(instruction, examples)
    return final_code


def getStreamlitCode(instruction, langchain_code, title):
    system_message_prompt = SystemMessagePromptTemplate.from_template(
        STREAMLIT_CODE_SYSTEM_TEMPLATE
    )
    human_message_prompt = HumanMessagePromptTemplate.from_template(
        STREAMLIT_CODE_HUMAN_TEMPLATE
    )
    chat_prompt = ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )
    chain = LLMChain(llm=llm, prompt=chat_prompt)
    merged_code = chain.run(
        instruction=instruction, langchain_code=langchain_code, title=title
    )
    merged_code = refineCode(merged_code)
    return merged_code


def get(
    instruction="Create a translation system that converts English to French",
    title="my translator",
    iterations=10,
):
    langchain_code = getLangChainCode(instruction, iterations=iterations)
    print(colored(langchain_code, "green"))
    return getStreamlitCode(instruction, langchain_code, title)


def streamlit_test():
    langchain_code = """
    from langchain.llms import OpenAI
    from langchain.chains import ConversationChain
    from langchain.memory import ConversationSummaryBufferMemory

    llm = OpenAI()

    psychologist = ConversationChain(
        llm=llm,
        memory=ConversationSummaryBufferMemory(llm=llm, max_token_limit=500),
        verbose=True
    )

    print("Welcome to the Personal Psychologist!")
    print("You can start by telling me about your problems or concerns.")

    while True:
        user_input = input("You: ")
        response = psychologist.predict(input=user_input)
        print("Psychologist:", response)
    """

    instruction = (
        "Create a personal psychologist that can remember the conversation history"
    )
    title = "My Psychologist"

    system_message_prompt = SystemMessagePromptTemplate.from_template(
        STREAMLIT_CODE_SYSTEM_TEMPLATE
    )
    human_message_prompt = HumanMessagePromptTemplate.from_template(
        STREAMLIT_CODE_HUMAN_TEMPLATE
    )
    chat_prompt = ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )
    chain = LLMChain(llm=llm, prompt=chat_prompt)
    merged_code = chain.run(
        instruction=instruction, langchain_code=langchain_code, title=title
    )
    merged_code = refineCode(merged_code)
    return merged_code


if __name__ == "__main__":
    fire.Fire(get)



from langchain import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.prompts.chat import (ChatPromptTemplate,
                                    HumanMessagePromptTemplate,
                                    SystemMessagePromptTemplate)

from chains.prompts import *


class Chains:
    llm = ChatOpenAI(model="gpt-3.5-turbo-16k", temperature=0)

    @classmethod
    def setLlm(cls, model, openai_api_key, temperature=0):
        cls.llm = ChatOpenAI(
            model=model, openai_api_key=openai_api_key, temperature=temperature
        )

    @classmethod
    def getChain(cls, system_template="", human_template="", **kwargs):
        prompts = []
        if system_template:
            prompts.append(SystemMessagePromptTemplate.from_template(system_template))
        if human_template:
            prompts.append(HumanMessagePromptTemplate.from_template(human_template))
        chat_prompt = ChatPromptTemplate.from_messages(prompts)
        return LLMChain(llm=cls.llm, prompt=chat_prompt).run(**kwargs)

    @classmethod
    def divide(cls, **kwargs):
        return cls.getChain(
            system_template=DIVIDE_TASKS_SYSTEM_TEMPLATE,
            human_template=DIVIDE_TASKS_HUMAN_TEMPLATE,
            **kwargs
        )

    @classmethod
    def merge(cls, **kwargs):
        return cls.getChain(
            system_template=MERGE_CODES_SYSTEM_TEMPLATE,
            human_template=MERGE_CODES_HUMAN_TEMPLATE,
            **kwargs
        )

    @classmethod
    def debug(cls, **kwargs):
        return cls.getChain(human_template=APP_DEBUGGING_TEMPLATE, **kwargs)

    @classmethod
    def draft(cls, **kwargs):
        return cls.getChain(human_template=DOC_USE_TEMPLATE, **kwargs)

    @classmethod
    def streamlit(cls, **kwargs):
        return cls.getChain(
            system_template=STREAMLIT_CODE_SYSTEM_TEMPLATE,
            human_template=STREAMLIT_CODE_HUMAN_TEMPLATE,
            **kwargs
        )

    @classmethod
    def feedback(cls, **kwargs):
        return cls.getChain(human_template=CODE_FEEDBACK_HUMAN_TEMPLATE, **kwargs)

    @classmethod
    def refine(cls, **kwargs):
        return cls.getChain(
            system_template=CODE_REFINE_SYSTEM_TEMPLATE,
            human_template=CODE_REFINE_HUMAN_TEMPLATE,
            **kwargs
        )

    @classmethod
    def refine1(cls, **kwargs):
        return cls.getChain(human_template=CODE_REFINE1_HUMAN_TEMPLATE, **kwargs)



DOC_USE_TEMPLATE = """
As a Python developer, your task is to use the LangChain library to develop an application based on a user-provided idea.

If the document is useful, incorporate the information from it into your Python code for the application.

Document: {document}

Application Idea: {idea}

Proceed with your Python code below:
"""

INIT_GENERATION_TEMPLATE = """
As a Python developer, your task is to use the LangChain library to develop an application based on a user-provided idea. A document is provided to assist you, but it may or may not be relevant.

If you find the document irrelevant to the application idea, respond with "Not relevant". If the document is useful, incorporate the information from it into your Python code for the application.

Document: {document}

Application Idea: {idea}

Proceed with your Python code below:
"""

APP_GENERATION_TEMPLATE = """
As a Python developer, your task is to create a new application with the LangChain library based on a user-specified idea. 
Reference successful examples in the library's repository, translate the user's idea into specific requirements, draft an architecture, develop the application, and validate your work against the examples, due to the one-shot nature of the task, attention to detail in understanding the examples, meticulous planning, and precise execution are paramount for successful application generation.

Possible answers:{possible_answers}

Application Idea:{idea}

LangChain Application Code:
"""

APP_DEBUGGING_TEMPLATE = """
As a software engineer, you are developing a Python application using the LangChain library. 
However, the current code is buggy and not functioning as expected. 
You have an application idea, the buggy code, and some feedback at your disposal. 
Analyze the buggy code in the context of the application idea, understand the issues based on the feedback received, propose necessary modifications, and implement the fixes to ensure the code aligns with the intended application idea.

Example Document:{document}
################################
Buggy Code:{draft_code}
################################
Application Idea:{idea}
################################
Feedback:{feedback}
################################
Refined Code:
"""

DIVIDE_TASKS_SYSTEM_TEMPLATE = """You are supposed to divide tasks into subtasks if they consist of more than one task otherwise, 
directly return the tasks itself"""

DIVIDE_TASKS_HUMAN_TEMPLATE = """
Think step by step.
First decide if the instruction consists of a single task.
if then directly return the task
Otherwise divide the instruction into list of multiple subtasks

Instruction: Generate a system which can generate a song from song title and give a score to that song out of 10
Subinstructions: ["generate a song from song title", "generate score from a song out of 10"]

Instruction: Create a translation system that converts English to French
Subinstructions: ["generate a English to French translator"]

Instruction: Create a system that generates tweet from hashtags and given keywords
Subinstructions: ["generate tweet from hashtags and keywords"]

Instruction: {task}
Subinstructions: 
"""

MERGE_CODES_SYSTEM_TEMPLATE = """You are supposed to create a final code to accomplish final task 
by looking at the code block which are created to do the subtasks.
Please merge those codes to do the final task"""


MERGE_CODES_HUMAN_TEMPLATE = """
{examples}
Final Task: {task}
Final Code:
"""


STREAMLIT_CODE_SYSTEM_TEMPLATE = """
    You are a helpful code assistant that can teach a junior developer how to code. Your language of choice is Python. 
    You will get a GOAL.
    You will get a logic code that does the logic part of the GOAL
    Your task is generating a stremlit code to add a UI to the logic to accomplish the GOAL 
    When needed, you can use streamlit's session_state methodology like in the below:
    ################################################
    Initialize values in Session State
    The Session State API follows a field-based API, which is very similar to Python dictionaries:

    # Initialization
    if 'key' not in st.session_state:
        streamlit.session_state['key'] = 'value'
    
    # You can use streamlit.session_state['any_key'] to store your objects which is supposed to store state variables
    """

STREAMLIT_CODE_HUMAN_TEMPLATE = """Write a streamlit code and add the logic code based on langchain library inside of it to accomplish the GOAL with the given title. Also when the result is ready add st.balloons() to the code 
Keep in mind that if the streamlit code includes a list which will be updated during the application, 
use st.session_state because for each update, it needs to be stored. 
Otherwise, the object will be resetted. 

GOAL: {instruction}
--------- 
Langchain based Logic Code:
{langchain_code} 
---------
Streamlit Application Title: {title}
Add st.title({title}) to the app
---------
Full Stremlit Code based on the langchain logic code:
"""
